"""
=============================================
Extracts PPG Cycles (:mod:`pypg.cycles`)
=============================================

This module is intended to extract PPG cycles and their characteristics.

Cycles
----------
find_onset          - Find the cycle(s) onset.
find_with_template  - Return valid cycle(s) based on a template.

References
----------
Elgendi, M., Norton, I., Brearley, M., Abbott, D., & Schuurmans, D.
(2013). Systolic Peak Detection in Acceleration Photoplethysmograms
Measured from Emergency Responders in Tropical Conditions. PLoS ONE,
8(10), 1–11. https://doi.org/10.1371/journal.pone.0076585

Li, Q., & Clifford, G. D. (2012). Dynamic time warping and machine
learning for signal quality assessment of pulsatile signals.
Physiological Measurement, 33(9), 1491–1501.
https://doi.org/10.1088/0967-3334/33/9/1491
"""

import math

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import signal, stats

from .plots import simple_plot, marks_plot


def find_onset(ppg, sampling_frequency, factor=0.667, distance=None, height=None,
                threshold=None, prominence=None, width=None, wlen=None,
                rel_height=0.5, plateau_size=None, verbose=False):
    """
    Finds the local minimun/minima that correspond to the onsets/start of the
    cardiac cycle(s).

    Parameters
    ----------
    ppg : pandas.Series, ndarray
        The PPG signal.
    sampling_frequency : int
        The sampling frequency of the signal in Hz.
    factor: float, optional
        Number that is used to calculate the distance in relation to the
        sampling_frequency, by default 0.667 (or 66.7%). The factor is based
        on the paper by Elgendi et al. (2013).
    distance : number, optional
        Minimum horizontal distance (>=1) between the cycles start points,
        by default None. However, the function assumes (factor *
        sampling_frequency) when None is given. For more information check
        the SciPy documentation.
    height : number or ndarray or sequence, optional
        Required height of peaks. Either a number, None, an array matching x
        or a 2-element sequence of the former, by default None. For more information check
        the SciPy documentation.
    threshold : number or ndarray or sequence, optional
        Required threshold of peaks, the vertical distance to its
        neighboring samples. Either a number, None, an array matching x or a
        2-element sequence of the former, by default None. For more
        information check the SciPy documentation.
    prominence : number or ndarray or sequence, optional
        Required prominence of peaks. Either a number, None, an array
        matching x or a 2-element sequence of the former, by default None.
        For more information check the SciPy documentation.
    width : number or ndarray or sequence, optional
        Required width of peaks in samples. Either a number, None, an array
        matching x or a 2-element sequence of the former, by default None.
        For more information check the SciPy documentation.
    wlen : int, optional
        Used for calculation of the peaks prominences, thus it is only used
        if one of the arguments prominence or width is given, by default
        None. For more information check the SciPy documentation.
    rel_height : float, optional
        Used for calculation of the peaks width, thus it is only used if
        width is given, by default 0.5 as defined by SciPy, by
        default None. For more information check the SciPy documentation.
    plateau_size : number or ndarray or sequence, optional
        Required size of the flat top of peaks in samples. Either a number,
        None, an array matching x or a 2-element sequence of the former, by
        default None. For more information check the SciPy documentation.
    verbose : boolean, optinal
        Verbose is used to print a graph with the PPG signal and the values
        of each cycle onset, by default False.

    Returns
    -------
    minimum/minima: np.ndarray
        Indice(s) value(s) in x with the start/onset of the PPG cycle(s).

    Raises
    ----------
    Exception
        When PPG values are neither pandas.Series nor ndarray.

    References
    ----------
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.find_peaks.html

    Elgendi, M., Norton, I., Brearley, M., Abbott, D., & Schuurmans, D.
    (2013). Systolic Peak Detection in Acceleration Photoplethysmograms
    Measured from Emergency Responders in Tropical Conditions. PLoS ONE,
    8(10), 1–11. https://doi.org/10.1371/journal.pone.0076585
    """

    if isinstance(ppg, pd.core.series.Series):
        signal_values = ppg.values
    elif isinstance(ppg, np.ndarray):
        signal_values = ppg
    else:
        raise Exception('PPG values not accepted, enter either'
                        +' pandas.Series or ndarray.')

    if distance is None:
        distance=factor*sampling_frequency

    # invert the signal to find the minimas
    peaks_data = signal.find_peaks(-signal_values, distance=distance, height=height,
                                threshold=threshold, prominence=prominence, width=width,
                                wlen=wlen, rel_height=rel_height, plateau_size=plateau_size)
    minima = peaks_data[0]
    if verbose:
        marks_plot(signal_values, minima)
    return minima

def find_with_template(ppg, sampling_frequency, return_type='original', factor=0.667,
                      distance=None, height=None, threshold=None, prominence=None, width=None,
                      wlen=None, rel_height=0.5, plateau_size=None, correlation_threshold=0.8,
                      verbose=False):
    """
    Finds PPG cycles in a PPG segment based on the method suggested by Li and
    Clifford (2012). All detected cycles in the same window are combined into a
    custom PPG signal template. Individual cycles are then compared with the
    template using two signal quality indices (SQI): (1) direct linear
    correlation and (2) direct linear correlation between the cycle, re-sampled
    to match the template length, and the template itself. Only if both
    correlations lie above a threshold, the cycle is considered valid.

    Parameters
    ----------
    ppg : pandas.Series, ndarray
        The PPG signal.
    sampling_frequency : int
        The sampling frequency of the signal in Hz.
    return_type: strm, optional
        The type of values to be returned (original or index), by default
        "original". Original returns a list of pd.Series or np.ndarray with the
        original data partioned into cycles. Index returns a list of tuples with
        the indexes of the begining and ending of each cycle in the original data.
    factor: float, optional
        Number that is used to calculate the distance in relation to the
        sampling_frequency, by default 0.667 (or 66.7%). The factor is based
        on the paper by Elgendi et al. (2013).
    distance : number, optional
        Minimum horizontal distance (>=1) between the cycles start points,
        by default None. However, the function assumes (factor *
        sampling_frequency) when None is given. For more information check
        the SciPy documentation.
    height : number or ndarray or sequence, optional
        Required height of peaks. Either a number, None, an array matching x
        or a 2-element sequence of the former, by default None. For more information check
        the SciPy documentation.
    threshold : number or ndarray or sequence, optional
        Required threshold of peaks, the vertical distance to its
        neighboring samples. Either a number, None, an array matching x or a
        2-element sequence of the former, by default None. For more
        information check the SciPy documentation.
    prominence : number or ndarray or sequence, optional
        Required prominence of peaks. Either a number, None, an array
        matching x or a 2-element sequence of the former, by default None.
        For more information check the SciPy documentation.
    width : number or ndarray or sequence, optional
        Required width of peaks in samples. Either a number, None, an array
        matching x or a 2-element sequence of the former, by default None.
        For more information check the SciPy documentation.
    wlen : int, optional
        Used for calculation of the peaks prominences, thus it is only used
        if one of the arguments prominence or width is given, by default
        None. For more information check the SciPy documentation.
    rel_height : float, optional
        Used for calculation of the peaks width, thus it is only used if
        width is given, by default 0.5 as defined by SciPy, by
        default None. For more information check the SciPy documentation.
    plateau_size : number or ndarray or sequence, optional
        Required size of the flat top of peaks in samples. Either a number,
        None, an array matching x or a 2-element sequence of the former, by
        default None. For more information check the SciPy documentation.
    correlation_threshold : float, optional
        Number that is used to calculate the correlation threshold from the
        template to the individual cycles, by default 0.8.
    verbose : boolean, optinal
        Verbose is used to print different graphs such as the onset of each
        cycle, the cycle template and all valid cycles, by default False.

    Raises
    ----------
    Exception
        When PPG values are neither pandas.Series nor ndarray.
        When return_type is neither 'original' or 'index'.

    Returns
    ----------
    cycles: list
        If "original" returns a list of pd.Series or np.ndarray with the valid PPG cycles.
        If "index" returns a list of tuples with the index of begining and
        ending of each valid PPG cycle.

    References
    ----------
    Li, Q., & Clifford, G. D. (2012). Dynamic time warping and machine
    learning for signal quality assessment of pulsatile signals.
    Physiological Measurement, 33(9), 1491–1501.
    https://doi.org/10.1088/0967-3334/33/9/1491
    """

    if isinstance(ppg, pd.core.series.Series):
        signal_values = ppg.values
    elif isinstance(ppg, np.ndarray):
        signal_values = ppg
    else:
        raise Exception('PPG values not accepted, enter either'
                        +' pandas.Series or ndarray.')

    if return_type not in ['index', 'original']:
        raise Exception('Wrong value for return_type.')

    initial_cycle_starts = find_onset(signal_values, sampling_frequency, factor, distance,
                                      height, threshold, prominence, width, wlen, rel_height,
                                      plateau_size, verbose)

    if len(initial_cycle_starts) <= 1:
        return []
    template_length = math.floor(np.median(np.diff(initial_cycle_starts)))
    cycle_starts = initial_cycle_starts[:-1]
    while cycle_starts[-1] + template_length > len(signal_values):
        cycle_starts = cycle_starts[:-1]
    template = []
    for i in range(template_length):
        template.append(np.mean(signal_values[cycle_starts + i]))

    corr_coef = []
    for cycle_start in cycle_starts:
        corr_coef.append(np.corrcoef(template,
                            signal_values[cycle_start:cycle_start+template_length])[0,1])

    valid_indices = np.argwhere(np.array(corr_coef) >= correlation_threshold)
    if (len(valid_indices) > len(cycle_starts) / 2) and len(valid_indices) > 1:
        cycle_starts = cycle_starts[np.squeeze(valid_indices)]
        template2 = []
        for i in range(template_length):
            template2.append(np.mean(signal_values[cycle_starts + i]))
        template = template2

    if verbose:
        simple_plot(template, title='Cycle Template')

    # check correlation of cycles with template SQI1: Pearson Correlation
    sqi1_corr = []
    for cycle_start in cycle_starts:
        corr, _ = stats.pearsonr(template, signal_values[cycle_start:cycle_start+template_length])
        sqi1_corr.append(corr)

    # SQI2: Pearson Correlation between the cycle, re-sampled to match the
    # template length, and the template itself
    sqi2_corr = []
    for cycle_start in cycle_starts:
        cycle_end = initial_cycle_starts[np.squeeze(
                                            np.argwhere(initial_cycle_starts==cycle_start)) + 1]
        corr, _ = stats.pearsonr(template, signal.resample(
                                    signal_values[cycle_start:cycle_end], template_length))
        sqi2_corr.append(corr)

    # filter for correlation >= correlation_threshold
    corrs = np.array([sqi1_corr, sqi2_corr]).transpose()
    cycle_starts = cycle_starts[np.all(corrs >= correlation_threshold, axis=1)]

    if verbose:
        print('Valid Cycles Detected')
        for cycle_start in cycle_starts:
            plt.plot(signal_values[cycle_start:cycle_start+template_length])
        plt.show()
        plt.close()

    cycles_indices = []
    for cycle_start in cycle_starts:
        cycle_end = initial_cycle_starts[np.squeeze(
                                            np.argwhere(initial_cycle_starts==cycle_start)) + 1]
        if (cycle_end - cycle_start) > template_length*1.2:
            cycle_end = cycle_start + template_length
        cycles_indices.append((cycle_start, cycle_end))

    if return_type == 'original':
        cycles = []
        for i, cycle_index in enumerate(cycles_indices):
            if isinstance(ppg, pd.core.series.Series):
                cycles.append(ppg.iloc[cycle_index[0]:cycle_index[1]])
            elif isinstance(ppg, np.ndarray):
                cycles.append(ppg[cycle_index[0]:cycle_index[1]])
        return cycles
    if return_type == 'index':
        return cycles_indices



"""
PyPG: A library for PPG (PhotoPlethysmoGram) processing


Documentation
-----------
Available in the docstrings for now; online soon.


Modules
-----------
Using the modules requires an explicit import. For example:
``import pypg.cycles``

    ::

    cycles
    features
    filters
    plots


Inspired By
-----------
SciPy: A scientific computing package for Python


References
----------
https://github.com/scipy/scipy

"""

import pypg.cycles
import pypg.features
import pypg.filters
import pypg.plots



"""
=============================================
Plots for PPG Signals (:mod:`pypg.plots`)
=============================================

This module implements a few plots for PPG signals.

Plots
----------
simple_plot - Plots a PPG signal.
marks_plot  - Plots a PPG signal with marks (e. g. peaks or valleys).

"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def _configure_plot(title=None, y_axis=None, x_axis=None):
    """
    Configures the plot colors, axis, font sizes, etc.

    Parameters
    ----------
    title : str, optional
        Title of the plot, by default None.
    y_axis : str, optional
        Y axis name, by default None.
    x_axis : str, optional
        X axis name, by default None.
    """
    # defines the parameters of the figure
    plt.rcParams.update({'font.size': 12})
    plt.rcParams['axes.edgecolor'] = '#333F4B'
    plt.rcParams['axes.linewidth'] = 0.8
    plt.rcParams['xtick.color'] = '#333F4B'
    plt.rcParams['ytick.color'] = '#333F4B'

    # changes the style of the axis spines
    _, axis = plt.subplots(figsize=(10, 5))
    axis.spines['top'].set_color('none')
    axis.spines['right'].set_color('none')
    axis.yaxis.grid(color='#333F4B', linestyle=':', linewidth=0.2, which='major')

    if title:
        axis.set_title(title, fontsize=16)
    if y_axis:
        axis.set_ylabel(y_axis, fontsize=14)
    if x_axis:
        axis.set_xlabel(x_axis, fontsize=14)

def simple_plot(ppg, title='PPG Signal', label='PPG Signal', y_axis='Amplitude',
                x_axis=None, figure_path=None):
    """
    Plots PPG signals.

    Parameters
    ----------
    ppg : indexable objects are supported.
        The PPG signal.
    title : str, optional
        Title of the plot, by default 'PPG Signal'.
    label : str, optional
        Label of the plotted data, by default 'PPG Signal'.
    y_axis : str, optional
        Y axis name, by default 'Amplitude'.
    x_axis : str, optional
        X axis name, by default None.
    figure_path : str, optional
        The path for the plot to be saved in pdf, by default None.
    """
    _configure_plot(title, y_axis, x_axis)
    plt.plot(ppg, label=label, color='#e8335e')
    plt.legend()
    if figure_path:
        plt.savefig(figure_path, dpi=300, bbox_inches='tight', format="pdf")
    plt.show()
    plt.close()

def marks_plot(ppg, marks, title='PPG Signal with Marks', label_ppg='PPG Signal',
               label_marks='Marks', y_axis='Amplitude', x_axis=None, figure_path=None):
    """
    Plots PPG signals with marks.

    Parameters
    ----------
    ppg : indexable objects are supported.
        The PPG signal.
    marks : ndarray
        Marks to be plotted against the PPG signal.
    title : str, optional
        Title of the plot, by default 'PPG Signal'.
    label_ppg : str, optional
        Label for the PPG signal, by default 'PPG Signal'.
    label_marks : str, optional
        Label for the marks, by default 'Marks'.
    y_axis : str, optional
        Y axis name, by default 'Amplitude'.
    x_axis : str, optional
        X axis name, by default None.
    figure_path : str, optional
        The path for the plot to be saved in pdf, by default None.
    """
    _configure_plot(title, y_axis, x_axis)

    plt.plot(ppg, color='#e8335e')
    for mark in marks:
        if isinstance(ppg, pd.core.series.Series):
            plt.plot(ppg.index[mark], ppg.iloc[mark], marker='X', markersize=8, color='#6233E8')
        elif isinstance(ppg, np.ndarray):
            plt.plot(mark, ppg[mark], marker='X', markersize=8, color='#6233E8')
    plt.legend([label_ppg, label_marks])
    if figure_path:
        plt.savefig(figure_path, dpi=300, bbox_inches='tight', format="pdf")
    plt.show()
    plt.close()



"""
=============================================
Extracts Features from PPG segments and cycles (:mod:`pypg.features`)
=============================================

This module is intended to extract features from PPG segments and cycles.

Features
----------
time       - Extract features from the time domain for a PPG segment.
time_cycle - Extract features from the time domain for a PPG cycle.

References
----------
Kurylyak, Y., Lamonaca, F., & Grimaldi, D. (2013). A Neural Network-based
method for continuous blood pressure estimation from a PPG signal.
Conference Record - IEEE Instrumentation and Measurement Technology
Conference, 280–283. https://doi.org/10.1109/I2MTC.2013.6555424

Elgendi, M., Norton, I., Brearley, M., Abbott, D., & Schuurmans, D.
(2013). Systolic Peak Detection in Acceleration Photoplethysmograms
Measured from Emergency Responders in Tropical Conditions. PLoS ONE,
8(10), 1–11. https://doi.org/10.1371/journal.pone.0076585

Li, Q., & Clifford, G. D. (2012). Dynamic time warping and machine
learning for signal quality assessment of pulsatile signals.
Physiological Measurement, 33(9), 1491–1501.
https://doi.org/10.1088/0967-3334/33/9/1491
"""

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import signal

from .cycles import find_with_template

def time(ppg, sampling_frequency, factor=0.6667, unit='ms', verbose=False):
    """
    Extracts time domain features from PPG cycles in a give PPG segment as
    described by Kurylyak et al. (2013). Returns a pandas.DataFrame in
    which each line contains the features for a given valid cycle (as described
    by Li et al.) from the PPG segment given as input.

    Parameters
    ----------
    ppg : pandas.Series, ndarray
        The PPG signal.
    sampling_frequency : int
        The sampling frequency of the signal in Hz.
    factor: float, optional
        Number that is used to calculate the distance in relation to the
        sampling_frequency, by default 0.667 (or 66.7%). The factor is based
        on the paper by Elgendi et al. (2013).
    unit : str, optional
        The unit of the index, by default 'ms'.
    verbose : bool, optional
        If True, plots the features with and without outliers, by default False.

    Raises
    ----------
    Exception
        When PPG values are neither pandas.Series nor ndarray.

    Returns
    -------
    segment_features : pd.DataFrame
        A dataframe with the time-domain features in seconds for each valid
        cycle in the PPG segment.
    """
    if isinstance(ppg, np.ndarray):
        ppg = pd.Series(ppg)
    elif not isinstance(ppg, pd.core.series.Series):
        raise Exception('PPG values not accepted, enter a pandas.Series or ndarray.')

    cycles = find_with_template(ppg, sampling_frequency, factor=0.6667, verbose=verbose)
    # all signal peaks
    all_peaks = signal.find_peaks(ppg.values, distance=factor*sampling_frequency)[0]
    if verbose:
        print('All Peaks: ', all_peaks)
    if len(cycles) == 0:
        return pd.DataFrame()

    cur_index = 0
    segment_features = pd.DataFrame()
    for i, cycle in enumerate(cycles):
        segment_features = segment_features.append(time_cycle(cycle,
                                         sampling_frequency, factor, unit, verbose=verbose),
                                         ignore_index=True)
        if i > 0:
            segment_features.loc[cur_index-1, 'CP'] = (
                segment_features.loc[cur_index, 'sys_peak_ts'] - segment_features.loc[cur_index-1,
                'sys_peak_ts']).total_seconds()
        cur_index = cur_index + 1
    # last cycle or only cycle need to relies on the difference between the general peaks
    all_peaks_index = len(all_peaks)-1
    segment_features.loc[cur_index-1, 'CP'] = (
        all_peaks[all_peaks_index] - all_peaks[all_peaks_index-1])/sampling_frequency

    if verbose:
        print('Cycle Features within Segment:')
        print(segment_features)

    # remove outliers
    segment_features = _clean_segment_features_of_outliers(segment_features)

    if verbose:
        print('Cycle Features within Segment and no Outliers:')
    return segment_features

def time_cycle(ppg, sampling_frequency, factor=0.667, unit='ms', verbose=False):
    """
    Extracts time domain features for a PPG cycle. Returns a pandas.Series with
    the features described by Kurylyak et al. (2013).

    Parameters
    ----------
    ppg : pandas.Series, ndarray
        The PPG signal.
    sampling_frequency : int
        The sampling frequency of the signal in Hz.
    factor: float, optional
        Number that is used to calculate the distance in relation to the
        sampling_frequency, by default 0.667 (or 66.7%). The factor is based
        on the paper by Elgendi et al. (2013).
    unit : str, optional
        The unit of the index, by default 'ms'.
    verbose : bool, optional
        If True, plots the features with and without outliers, by default False.

    Raises
    ----------
    Exception
        When PPG values are neither pandas.Series nor ndarray.

    Returns
    -------
    cycle_features : pd.DataFrame
        A dataframe with the time-domain features in seconds for the PPG cycle.
    """

    if isinstance(ppg, np.ndarray):
        ppg = pd.Series(ppg)
    elif not isinstance(ppg, pd.core.series.Series):
        raise Exception('PPG values not accepted, enter a pandas.Series or ndarray.')

    if not isinstance(ppg.index, pd.DatetimeIndex):
        ppg.index = pd.to_datetime(ppg.index, unit=unit)

    ppg = ppg.interpolate(method='time')
    ppg = ppg - ppg.min()
    max_amplitude = ppg.max()

    peaks = signal.find_peaks(ppg.values, distance=factor*sampling_frequency)[0]
    sys_peak_ts = ppg.index[peaks[0]]

    if verbose:
        plt.figure()
        plt.xlim((ppg.index.min(), ppg.index.max()))
        plt.scatter(ppg.index[peaks], ppg[peaks])
        plt.plot(ppg.index, ppg.values)

    # features
    cycle_features = pd.DataFrame({
                        'start_ts': ppg.index.min(),
                        'sys_peak_ts': sys_peak_ts,
                        'SUT': (sys_peak_ts - ppg.index.min()).total_seconds(),
                        'DT': (ppg.index.max() - sys_peak_ts).total_seconds()
                        }, index=[0])
    for p_value in [10, 25, 33, 50, 66, 75]:
        p_ampl = p_value / 100 * max_amplitude
        x_1, x_2 = _find_xs_for_y(ppg, p_ampl, peaks[0])
        if verbose:
            plt.scatter([x_1, x_2], ppg[[x_1, x_2]])
        cycle_features.loc[0, 'DW_'+str(p_value)] = (x_2 - sys_peak_ts).total_seconds()
        cycle_features.loc[0, 'DW_SW_sum_'+str(p_value)] = (x_2 - x_1).total_seconds()
        cycle_features.loc[0, 'DW_SW_ratio_'+str(p_value)] = (
                                        x_2 - sys_peak_ts) / (sys_peak_ts - x_1)
    if verbose:
        plt.show()
    return cycle_features

# takes a dataframe of calculated features and removes the outliers occurring due
# to inaccuracies in the signal
def _clean_segment_features_of_outliers(segment_df, treshold=0.8):
    quant = segment_df.quantile(treshold)
    for col in segment_df.columns:
        if col.find('ts') == -1:
            segment_df = segment_df[segment_df[col] < quant[col]*2]
    return segment_df

# returns the x values for those samples in the signal, that are closest to some given y value
def _find_xs_for_y(y_s, y_val, sys_peak):
    diffs = abs(y_s - y_val)
    x_1 = diffs[:sys_peak].idxmin()
    x_2 = diffs[sys_peak:].idxmin()
    return x_1, x_2



"""
=============================================
Filters for PPG Signals (:mod:`pypg.filters`)
=============================================

This module implements the most interesting filters found in the literature for
PPG signals.

Filters
----------
butterfy - Butterworth filter optimized for PPG signals.
chebyfy  - Chebyshev II filter optimized for PPG signals.
movefy   - Moving average filter.

Returns
----------
filtered : pandas.Series or ndarray
    Filtered PPG signal.

References
----------
Liang, Y., Elgendi, M., Chen, Z., & Ward, R. (2018). An optimal filter for
short photoplethysmogram signals. Scientific Data, 5(1), 180076.
https://doi.org/10.1038/sdata.2018.76

"""

import numpy as np
import pandas as pd
from scipy import signal

from .plots import simple_plot


def butterfy(ppg, cutoff_frequencies, sampling_frequency, filter_type='bandpass',
             filter_order=4, verbose=False, figure_path=None):
    """
    Implements the Butterworth filter with 'bandpass' as the filter_type and filter_order of
    4. Described by Liang et al. (2018) as the second best option for filtering PPG
    signals.

    Parameters
    ----------
    ppg : pandas.Series, ndarray
        The PPG signal.
    cutoff_frequencies : int or list
        The cutoff frequency(ies if bandpass) for the filter in Hz.
    sampling_frequency : int
        The sampling frequency of the signal in Hz.
    filter_type : str, optional
        Filter type (lowpass, highpass, bandpass). Defaults to 'bandpass'.
    filter_order : int, optional
        Filter Order. Defaults to 4.
    verbose : bool, optional
        Plots the signal. Defaults to False.
    figure_path : str, optional
        The path for the PPG plot to be saved.

    Raises
    ----------
    Exception
        When PPG values are neither pandas.Series nor ndarray.

    Returns
    ----------
    filtered: pandas.Series or ndarray
        Filtered PPG signal.

    References
    ----------
    Liang, Y., Elgendi, M., Chen, Z., & Ward, R. (2018). An optimal filter
    for short photoplethysmogram signals. Scientific Data, 5(1), 180076.
    https://doi.org/10.1038/sdata.2018.76
    """
    if isinstance(ppg, pd.core.series.Series):
        signal_values = ppg.values
        signal_index = ppg.index
    elif isinstance(ppg, np.ndarray):
        signal_values = ppg
    else:
        raise Exception('PPG values not accepted, enter either'
                        +' pandas.Series or ndarray.')

    sos = signal.butter(filter_order, cutoff_frequencies, btype=filter_type,
                        fs=sampling_frequency, output='sos')
    filtered = signal.sosfiltfilt(sos, signal_values)

    if isinstance(ppg, pd.core.series.Series):
        filtered = pd.Series(filtered)
        filtered.index = signal_index

    if verbose:
        if isinstance(cutoff_frequencies, list):
            cut = ', '.join(str(s) for s in cutoff_frequencies)
            cut = ' ['+cut+']'
        elif isinstance(cutoff_frequencies, int):
            cut = str(cutoff_frequencies)
        label = 'Buterworth'+' '+filter_type+' '+'filter '+cut+' Hz'
        simple_plot(ppg=filtered, title='Filtered Signal', label=label, figure_path=figure_path)
    return filtered

def chebyfy(ppg, cutoff_frequencies, sampling_frequency, filter_type='bandpass',
            filter_order=4, band_attenuation=20, verbose=False, figure_path=None):
    """
    Imlements the Chebyshev II filter with 'bandpass' as the filter_type, a filter_order of 4,
    and a band_attenuation of 20 dB. Described by Liang et al. (2018) as the best filter for
    PPG signals.

    Parameters
    ----------
    ppg : pandas.Series or ndarray
            The PPG signal.
    cutoff_frequencies : int or list
            The cutoff frequency(ies if bandpass) for the filter in Hz.
    sampling_frequency : int
            The sampling frequency of the signal in Hz.
    filter_type : str, optional
            Filter type (lowpass, highpass, bandpass). Defaults to 'bandpass'.
    filter_order : int, optional
            Filter Order. Defaults to 4.
    band_attenuation : int, optional
            The attenuation required in the stop band.
            In decibels (dB), as a positive number. Defaults to 20.
    verbose : bool, optional
            Plots the signal. Defaults to False.
    figure_path : str, optional
            The path for the PPG plot to be saved.

    Raises
    ----------
    Exception
        When PPG values are neither pandas.Series nor ndarray.

    Returns
    ----------
    filtered: pandas.Series or ndarray
    Filtered PPG signal.

    References
    ----------
    Liang, Y., Elgendi, M., Chen, Z., & Ward, R. (2018). An optimal filter for
    short photoplethysmogram signals. Scientific Data, 5(1), 180076.
    https://doi.org/10.1038/sdata.2018.76
    """
    if isinstance(ppg, pd.core.series.Series):
        signal_values = ppg.values
        signal_index = ppg.index
    elif isinstance(ppg, np.ndarray):
        signal_values = ppg
    else:
        raise Exception('PPG values not accepted, enter either'
                        +' pandas.Series or ndarray.')

    sos = signal.cheby2(filter_order, band_attenuation, cutoff_frequencies,
                        btype=filter_type, fs=sampling_frequency, output='sos')
    filtered = signal.sosfiltfilt(sos, signal_values)

    if isinstance(ppg, pd.core.series.Series):
        filtered = pd.Series(filtered)
        filtered.index = signal_index

    if verbose:
        if isinstance(cutoff_frequencies, list):
            cut = ', '.join(str(s) for s in cutoff_frequencies)
            cut = ' ['+cut+']'
        elif isinstance(cutoff_frequencies, int):
            cut = str(cutoff_frequencies)
        label = 'Chebyshev II'+' '+filter_type+' '+'filter '+cut+' Hz'
        simple_plot(ppg=filtered, title='Filtered Signal', label=label, figure_path=figure_path)
    return filtered

def movefy(ppg, size, verbose=False, figure_path=None):
    """
    Implements the moving average filter.

    Parameters
    ----------
    ppg : pandas.Series or ndarray
        The PPG signal.
    size : int
        Size of the np.ones mask to be convoluted.
    verbose : bool, optional
        Plot the signal. Defaults to False, by default False.
    figure_path : [type], optional
        The path for the PPG plot to be saved, by default None.

    Returns
    -------
    filtered: pandas.Series or ndarray
        Filtered PPG signal.

    Raises
    ------
    Exception
        When PPG values are neither pandas.Series nor ndarray.

    References
    ----------
    https://stackoverflow.com/questions/13728392/moving-average-or-running-mean/27681394#27681394
    """
    if isinstance(ppg, pd.core.series.Series):
        signal_values = ppg.values
        signal_index = ppg.index
    elif isinstance(ppg, np.ndarray):
        signal_values = ppg
    else:
        raise Exception('PPG values not accepted, enter either'
                        +' pandas.Series or ndarray.')

    filtered = signal.convolve(signal_values, np.ones((size,))/size, mode='valid')

    if isinstance(ppg, pd.core.series.Series):
        filtered = pd.Series(filtered)
        padding = pd.Series(np.nan, index=range(len(signal_index) - len(filtered.index)))
        filtered = filtered.append(padding,ignore_index=True)
        filtered.index = signal_index

    if verbose:
        simple_plot(ppg=filtered, title='Filtered Signal',
                    label='Moving average filter', figure_path=figure_path)
    return filtered
